#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

"""
File containing Series APIs defined in the Modin API layer, but with different behavior in Snowpark
pandas, such as `Series.memory_usage`.
"""

from typing import Union

import pandas as native_pd

from snowflake.snowpark.modin import pandas as pd  # noqa: F401
from snowflake.snowpark.modin.pandas import Series
from snowflake.snowpark.modin.pandas.api.extensions import register_series_accessor
from snowflake.snowpark.modin.plugin._internal.telemetry import (
    snowpark_pandas_telemetry_method_decorator,
)
from snowflake.snowpark.modin.plugin._typing import ListLike
from snowflake.snowpark.modin.plugin.utils.error_message import series_not_implemented
from snowflake.snowpark.modin.utils import _inherit_docstrings


@_inherit_docstrings(native_pd.Series.memory_usage, apilink="pandas.Series")
@register_series_accessor("memory_usage")
@snowpark_pandas_telemetry_method_decorator
def memory_usage(self, index: bool = True, deep: bool = False) -> int:
    """
    Return zero bytes for memory_usage
    """
    # TODO: SNOW-1264697: push implementation down to query compiler
    return 0


@_inherit_docstrings(native_pd.Series.infer_objects, apilink="pandas.Series")
@register_series_accessor("infer_objects")
@snowpark_pandas_telemetry_method_decorator
@series_not_implemented()
def infer_objects(self) -> Series:  # pragma: no cover # noqa: RT01, D200
    """
    Attempt to infer better dtypes for object columns.
    """
    return self.__constructor__(query_compiler=self._query_compiler.infer_objects())


@_inherit_docstrings(native_pd.Series.nunique, apilink="pandas.Series")
@register_series_accessor("nunique")
@snowpark_pandas_telemetry_method_decorator
def nunique(self, dropna: bool = True) -> int:
    """
    Return number of unique elements in the series.

    Excludes NA values by default. Snowpark pandas API does not distinguish between different NaN types like None,
    pd.NA or np.nan and treats them as the same

    Parameters
    ----------
    dropna : bool, default True
        Don't include NaN in the count.

    Returns
    -------
    int

    Examples
    --------
    >>> import snowflake.snowpark.modin.pandas as pd
    >>> import numpy as np
    >>> s = pd.Series([1, 3, 5, 7, 7])
    >>> s
    0    1
    1    3
    2    5
    3    7
    4    7
    dtype: int8

    >>> s.nunique()
    4

    >>> s = pd.Series([pd.NaT, np.nan, pd.NA, None, 1])
    >>> s.nunique()
    1

    >>> s.nunique(dropna=False)
    2

    """
    # TODO: SNOW-1264688: remove this override
    return super(Series, self).nunique(dropna=dropna)


@_inherit_docstrings(native_pd.Series.isin, apilink="pandas.Series")
@register_series_accessor("isin")
@snowpark_pandas_telemetry_method_decorator
def isin(self, values: Union[set, ListLike]) -> Series:
    """
    Whether elements in Series are contained in `values`.

    Return a boolean Series showing whether each element in the Series
    matches an element in the passed sequence of `values`.

    Caution
    -------
    Snowpark pandas deviates from pandas here with respect to NA values: when the value is considered NA or
    values contains at least one NA, None is returned instead of a boolean value.

    Parameters
    ----------
    values : set or list-like
        The sequence of values to test. Passing in a single string will
        raise a ``TypeError``. Instead, turn a single string into a
        list of one element.

    Returns
    -------
    Series
        Series of booleans indicating if each element is in values.

    Examples
    --------
    >>> s = pd.Series(['lama', 'cow', 'lama', 'beetle', 'lama',
    ...                'hippo'], name='animal')
    >>> s.isin(['cow', 'lama'])
    0     True
    1     True
    2     True
    3    False
    4     True
    5    False
    Name: animal, dtype: bool

    To invert the boolean values, use the ``~`` operator:

    >>> ~s.isin(['cow', 'lama'])
    0    False
    1    False
    2    False
    3     True
    4    False
    5     True
    Name: animal, dtype: bool

    Passing a single string as ``s.isin('lama')`` will raise an error. Use
    a list of one element instead:

    >>> s.isin(['lama'])
    0     True
    1    False
    2     True
    3    False
    4     True
    5    False
    Name: animal, dtype: bool

    >>> pd.Series([1]).isin(['1'])
    0    False
    dtype: bool

    >>> pd.Series([1, 2, None]).isin([2])
    0    False
    1     True
    2     None
    dtype: object
    """

    # pandas compatible TypeError
    if isinstance(values, str):
        raise TypeError(
            "only list-like objects are allowed to be passed to isin(), you passed a [str]"
        )

    # convert to list if given as set
    if isinstance(values, set):
        values = list(values)

    return super(Series, self).isin(values)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from collections.abc import Hashable
from typing import Literal, NamedTuple, Optional, Union

import numpy as np

# Snowpark pandas API always treats the pandas label as a tuple(LabelComponent), when the length of tuple is > 1,
# it represents multi-index, otherwise it is single level.
LabelComponent = Hashable
LabelTuple = tuple[LabelComponent, ...]
# can be removed once move to pandas 2.0
DropKeep = Literal["first", "last", False]

# pandas defines list-like as objects that are considered list-like are for example Python lists, tuples, sets, NumPy arrays,
# and pandas Series according to https://pandas.pydata.org/docs/reference/api/pandas.api.types.is_list_like.html. Define
# them for Snowpark pandas here. Note that we exclude Snowpark pandas Series here explicitly.
ListLike = Union[set, list, tuple, np.ndarray]

ListLikeOfFloats = Union[set[float], list[float], tuple[float]]


class LabelIdentifierPair(NamedTuple):
    """
    pair between pandas label and the corresponding snowflake quoted identifier.
    """

    # Internal representation for pandas label used to access pandas dataframe
    label: LabelTuple
    # Used to access the snowpark dataframe with data in snowflake
    snowflake_quoted_identifier: str


JoinTypeLit = Literal["left", "right", "inner", "outer", "cross"]
AlignTypeLit = Literal[
    # If align column values matches exactly, merge frames line by line (this is
    # equivalent to joining on row position) otherwise perform LEFT OUTER JOIN on
    # align columns.
    "left",
    # If align column values matches exactly, merge frames line by line (this is
    # equivalent to joining on row position) otherwise perform FULL OUTER JOIN on
    # align columns.
    "outer",
    # If align column values matches exactly, merge frames line by line (this is
    # equivalent to joining on row position) otherwise
    # - perform LEFT OUTER JOIN if left frame is non-empty
    # - perform RIGHT OUTER JOIN if left frame is empty
    "coalesce",
]  # right and inner can also be supported if needed

SnowflakeSupportedFileTypeLit = Union[
    Literal["csv"], Literal["json"], Literal["parquet"]
]


class PandasLabelToSnowflakeIdentifierPair(NamedTuple):
    """
    Pair between pandas label and the corresponding snowflake quoted identifier.
    """

    # pandas label
    pandas_label: Optional[Hashable]
    # Snowflake quoted identifier
    snowflake_quoted_identifier: str


# once updated to pandas 2.0, remove this, because this can be directly imported from pandas._typing
InterpolateOptions = Literal[
    "linear",
    "time",
    "index",
    "values",
    "nearest",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "barycentric",
    "polynomial",
    "krogh",
    "piecewise_polynomial",
    "spline",
    "pchip",
    "akima",
    "cubicspline",
    "from_derivatives",
]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import functools
from collections.abc import Hashable
from dataclasses import dataclass

from pandas._typing import Callable, Scalar

from snowflake.snowpark.column import Column as SnowparkColumn
from snowflake.snowpark.functions import col, concat, floor, iff, repeat, when
from snowflake.snowpark.modin.plugin._internal.frame import InternalFrame
from snowflake.snowpark.modin.plugin._internal.join_utils import (
    JoinOrAlignInternalFrameResult,
)
from snowflake.snowpark.modin.plugin._internal.type_utils import infer_object_type
from snowflake.snowpark.modin.plugin._internal.utils import pandas_lit
from snowflake.snowpark.modin.plugin.utils.error_message import ErrorMessage
from snowflake.snowpark.types import (
    DataType,
    NullType,
    StringType,
    _FractionalType,
    _IntegralType,
)

NAN_COLUMN = pandas_lit("nan").cast("float")

# set of supported binary operations that can be mapped to Snowflake
SUPPORTED_BINARY_OPERATIONS = {
    "truediv",
    "rtruediv",
    "floordiv",
    "rfloordiv",
    "mod",
    "rmod",
    "pow",
    "rpow",
    "__or__",
    "__ror__",
    "__and__",
    "__rand__",
    "add",
    "radd",
    "sub",
    "rsub",
    "mul",
    "rmul",
    "eq",
    "ne",
    "gt",
    "lt",
    "ge",
    "le",
}


def compute_modulo_between_snowpark_columns(
    first_operand: SnowparkColumn,
    first_datatype: DataType,
    second_operand: SnowparkColumn,
    second_datatype: DataType,
) -> SnowparkColumn:
    """
    Compute modulo between two Snowpark columns ``first_operand`` and ``second_operand``.
    Supports only numeric values for operands, raises NotImplementedError otherwise.
    Module may produce results different from native pandas or Python.
    """
    # 0. if f or s is NULL, return NULL (Snowflake's rule)
    # 1. s == 0, return nan
    # 2. if s != 0, return f % s
    #
    #     Examples
    # --------
    # >>> a = pd.Series([7, 7, -7, -7])
    # >>> b = pd.Series([5, -5, 5, -5])
    # >>> a % b
    # 0    2.0
    # 1    2.0
    # 2   -2.0
    # 3   -2.0
    # dtype: float64

    # >>> a = pd.Series([8.9, -0.22, np.nan, -1.02, 3.15, 2.0])
    # >>> b = pd.Series([-2.3, -76.34, 5.3, 5.3, 8.12])
    # >>> a % b
    # 0    2.00
    # 1   -0.22
    # 2     NaN
    # 3   -1.02
    # 4    3.15
    # 5     NaN
    # dtype: float64

    # Behavior differences
    # --------------------
    # Python               pandas 1.5            Snowflake
    #  7 %  5 =  2          7 %  5 =  2           7 %  5 =  2
    #  7 % -5 = -3          7 % -5 = -3           7 % -5 =  2
    # -7 %  5 =  3         -7 %  5 =  3          -7 %  5 = -2
    # -7 % -5 = -2         -7 % -5 = -2          -7 % -5 = -2
    #
    # Snowpark pandas API differs from native pandas results whenever an operand with a negative
    # sign is used.

    is_first_operand_numeric_type = (
        isinstance(first_datatype, _IntegralType)
        or isinstance(first_datatype, _FractionalType)
        or isinstance(first_datatype, NullType)
    )

    is_second_operand_numeric_type = (
        isinstance(second_datatype, _IntegralType)
        or isinstance(second_datatype, _FractionalType)
        or isinstance(second_datatype, NullType)
    )

    if is_first_operand_numeric_type and is_second_operand_numeric_type:
        return (
            when(first_operand.is_null() | second_operand.is_null(), None)
            .when(second_operand == 0, NAN_COLUMN)
            .otherwise(first_operand % second_operand)
        )
    else:
        ErrorMessage.not_implemented(
            "Modulo does not support non-numeric types, consider using a UDF with apply instead."
        )


def compute_power_between_snowpark_columns(
    first_operand: SnowparkColumn,
    second_operand: SnowparkColumn,
) -> SnowparkColumn:
    """
    Compute power between two Snowpark columns ``first_operand`` and ``second_operand``.
    """
    # 0. if f == 1 or s == 0, return 1 or 1.0 based on f's type (pandas' behavior)
    # 1. if f or s is NULL, return NULL (Snowflake's behavior)
    # 2. if f is nan, or s is nan, or f < 0 and s can not be cast to int without loss (int(s) != s), return nan
    #    In Snowflake, if f < 0 and s is not an integer, an invalid floating point operation will be raised.
    #    E.g., pow(-7, -10.0) is valid, but pow(-7, -10.1) is invalid in snowflake.
    #    In pandas, pow(-7, -10.1) returns NaN.
    # 3. else return f ** s
    result = (
        when((first_operand == 1) | (second_operand == 0), 1)
        .when(first_operand.is_null() | second_operand.is_null(), None)
        .when(
            (first_operand == NAN_COLUMN)
            | (second_operand == NAN_COLUMN)
            | (
                (first_operand < 0)
                # it checks whether the value can be cast int without loss
                & (second_operand.cast("int") != second_operand)
            ),
            NAN_COLUMN,
        )
        .otherwise(first_operand**second_operand)
    )
    return result


def is_binary_op_supported(op: str) -> bool:
    """
    check whether binary operation is mappable to Snowflake
    Args
        op: op as string

    Returns:
        True if binary operation can be mapped to Snowflake/Snowpark, else False
    """

    return op in SUPPORTED_BINARY_OPERATIONS


def compute_binary_op_between_snowpark_columns(
    op: str,
    first_operand: SnowparkColumn,
    first_datatype: Callable[[], DataType],
    second_operand: SnowparkColumn,
    second_datatype: Callable[[], DataType],
) -> SnowparkColumn:
    """
    Compute pandas binary operation for two SnowparkColumns
    Args:
        op: pandas operation
        first_operand: SnowparkColumn for lhs
        first_datatype: Callable for Snowpark Datatype for lhs, this is lazy so we can avoid pulling the value if
        it is not needed.
        second_operand: SnowparkColumn for rhs
        second_datatype: Callable for Snowpark DateType for rhs, this is lazy so we can avoid pulling the value if
        it is not needed.

    Returns:
        SnowparkColumn expr for translated pandas operation
    """

    binary_op_result_column = None

    # some operators and the data types have to be handled specially to align with pandas
    # However, it is difficult to fail early if the arithmetic operator is not compatible
    # with the data type, so we just let the server raise exception (e.g. a string minus a string).
    if op in ["truediv", "rtruediv", "floordiv", "rfloordiv"]:
        # rtruediv means b/a, rfloordiv means b//a in Python
        if op in ["rtruediv", "rfloordiv"]:
            first_operand, second_operand = (
                second_operand,
                first_operand,
            )

        binary_op_result_column = first_operand / second_operand

        if op in ["floordiv", "rfloordiv"]:
            binary_op_result_column = floor(binary_op_result_column)
    elif op in ["mod", "rmod"]:
        if op == "rmod":
            first_operand, second_operand = (
                second_operand,
                first_operand,
            )
        binary_op_result_column = compute_modulo_between_snowpark_columns(
            first_operand, first_datatype(), second_operand, second_datatype()
        )
    elif op in ["pow", "rpow"]:
        if op == "rpow":
            first_operand, second_operand = (
                second_operand,
                first_operand,
            )
        binary_op_result_column = compute_power_between_snowpark_columns(
            first_operand, second_operand
        )
    elif op in ["__or__", "__ror__"]:
        binary_op_result_column = first_operand | second_operand
    elif op in ["__and__", "__rand__"]:
        binary_op_result_column = first_operand & second_operand
    elif op in ["add", "radd", "mul", "rmul"]:

        # string/string case (only for add/radd)
        if isinstance(second_datatype(), StringType) and isinstance(
            first_datatype(), StringType
        ):
            if "add" == op:
                binary_op_result_column = concat(first_operand, second_operand)
            elif "radd" == op:
                binary_op_result_column = concat(second_operand, first_operand)

        # string/integer case (only for mul/rmul)
        if op in ["mul", "rmul"] and (
            (
                isinstance(second_datatype(), _IntegralType)
                and isinstance(first_datatype(), StringType)
            )
            or (
                isinstance(second_datatype(), StringType)
                and isinstance(first_datatype(), _IntegralType)
            )
        ):
            # Snowflake's repeat doesn't support negative number
            # but pandas will return an empty string

            # swap first_operand with second_operand because REPEAT(<input>, <n>) expects <input> to be string
            if isinstance(first_datatype(), _IntegralType):
                first_operand, second_operand = second_operand, first_operand

            binary_op_result_column = iff(
                second_operand > pandas_lit(0),
                repeat(first_operand, second_operand),
                pandas_lit(""),
            )

    # If there is no special binary_op_result_column result, it means the operator and
    # the data type of the column don't need special handling. Then we get the overloaded
    # operator from Snowpark Column class, e.g., __add__ to perform binary operations.
    if binary_op_result_column is None:
        binary_op_result_column = getattr(first_operand, f"__{op}__")(second_operand)

    return binary_op_result_column


def compute_binary_op_between_snowpark_column_and_scalar(
    op: str,
    first_operand: SnowparkColumn,
    datatype: Callable[[], DataType],
    second_operand: Scalar,
) -> SnowparkColumn:
    """
    Compute the binary operation between a Snowpark column and a scalar.
    Args:
        op: the name of binary operation
        first_operand: The SnowparkColumn for lhs
        datatype: Callable for Snowpark data type, this is lazy so we can avoid pulling the value if
        it is not needed.
        second_operand: Scalar value

    Returns:
        The result as a Snowpark column
    """

    def second_datatype() -> DataType:
        return infer_object_type(second_operand)

    second_operand = pandas_lit(second_operand)

    return compute_binary_op_between_snowpark_columns(
        op, first_operand, datatype, second_operand, second_datatype
    )


def compute_binary_op_between_scalar_and_snowpark_column(
    op: str,
    first_operand: Scalar,
    second_operand: SnowparkColumn,
    datatype: Callable[[], DataType],
) -> SnowparkColumn:
    """
    Compute the binary operation between a scalar and a Snowpark column.
    Args:
        op: the name of binary operation
        first_operand: Scalar value
        second_operand: The SnowparkColumn for rhs
        datatype: Callable for Snowpark data type, this is lazy so we can avoid pulling the value if
        it is not needed.

    Returns:
        The result as a Snowpark column
    """

    def first_datatype() -> DataType:
        return infer_object_type(first_operand)

    first_operand = pandas_lit(first_operand)

    return compute_binary_op_between_snowpark_columns(
        op, first_operand, first_datatype, second_operand, datatype
    )


def compute_binary_op_with_fill_value(
    op: str,
    lhs: SnowparkColumn,
    lhs_datatype: Callable[[], DataType],
    rhs: SnowparkColumn,
    rhs_datatype: Callable[[], DataType],
    fill_value: Scalar,
) -> SnowparkColumn:
    """
    Helper method for performing binary operations.
    1. Fills NaN/None values in the lhs and rhs with the given fill_value.
    2. Computes the binary operation expression for lhs <op> rhs.

    fill_value replaces NaN/None values when only either lhs or rhs is NaN/None, not both lhs and rhs.
    For instance, with fill_value = 100,
    1. Given lhs = None and rhs = 10, lhs is replaced with fill_value.
           result = lhs + rhs => None + 10 => 100 (replaced) + 10 = 110
    2. Given lhs = 3 and rhs = None, rhs is replaced with fill_value.
           result = lhs + rhs => 3 + None => 3 + 100 (replaced) = 103
    3. Given lhs = None and rhs = None, neither lhs nor rhs is replaced since they both are None.
           result = lhs + rhs => None + None => None.

    Args:
        op: pandas operation to perform between lhs and rhs
        lhs: the lhs SnowparkColumn
        lhs_datatype: Callable for Snowpark Datatype for lhs
        rhs: the rhs SnowparkColumn
        rhs_datatype: Callable for Snowpark Datatype for rhs
        fill_value: Fill existing missing (NaN) values, and any new element needed for
            successful DataFrame alignment, with this value before computation.

    Returns:
        SnowparkColumn expression for translated pandas operation
    """
    lhs_cond, rhs_cond = lhs, rhs
    if fill_value is not None:
        fill_value_lit = pandas_lit(fill_value)
        lhs_cond = iff(lhs.is_null() & ~rhs.is_null(), fill_value_lit, lhs)
        rhs_cond = iff(rhs.is_null() & ~lhs.is_null(), fill_value_lit, rhs)

    return compute_binary_op_between_snowpark_columns(
        op, lhs_cond, lhs_datatype, rhs_cond, rhs_datatype
    )


def merge_label_and_identifier_pairs(
    sorted_column_labels: list[str],
    q_frame_sorted: list[tuple[str, str]],
    q_missing_sorted: list[tuple[str, str]],
) -> list[tuple[str, str]]:
    """
    Helper function to create a merged list of column label/snowflake quoted identifiers. Assume q_frame_sorted and q_missing_sorted are disjoint wrt to labels.

    Example:
         Given sorted_column_labels = [1, 2, 3]
         and q_frame_sorted =  [(1, "A"), (3, "C")]    q_missing_sorted =  [(2, "B")]
         this function will produce as output [(1, "A"), (2, "B"), (3, "C")].
         Each q_frame_sorted and q_missing_sorted are lists of label/identifier pairs.
         I.e., [(1, "A"), (3, "C")] should be understood as 1 -> "A", 3 -> "B".
         They're each assumed to be sorted with respect to their labels, and all labels must be contained within
         the sorted_column_labels variable.
         The result is a combined, sorted representation 1 -> "A", 2 -> "B", 3 -> "C" which resembles the merge-step
         of a classical mergesort algorithm.
    Args:
        sorted_column_labels: The labels to merge for
        q_frame_sorted: sorted list of label/identifier pairs. All labels must be contained within sorted_column_labels.
        q_missing_sorted: sorted list of label/identifier pairs. All labels must be contained within sorted_column_labels.

    Returns:
        List of label/identifier pairs. If the labels were projected out, they would form sorted_column_labels.
    """
    if len(q_frame_sorted) > 0 and len(q_missing_sorted) > 0:
        # merge labels/identifiers

        i_frame = 0
        i_missing = 0

        pairs = []
        for label in sorted_column_labels:
            # Leave merge iff either queue is exhausted.
            if i_frame >= len(q_frame_sorted) or i_missing >= len(q_missing_sorted):
                break

            if label == q_frame_sorted[i_frame][0]:
                pairs.append(q_frame_sorted[i_frame])
                i_frame += 1
            elif label == q_missing_sorted[i_missing][0]:
                pairs.append(q_missing_sorted[i_missing])
                i_missing += 1
            # else case not relevant here, because labels of q_frame_sorted and q_missing_sorted must be disjoint.

        if i_frame < len(q_frame_sorted):
            pairs += q_frame_sorted[i_frame:]
        elif i_missing < len(q_missing_sorted):
            pairs += q_missing_sorted[i_missing:]

        return pairs
    elif len(q_missing_sorted) == 0:
        return q_frame_sorted
    else:
        return q_missing_sorted


@dataclass
class BinaryOperationPair:
    # For detailed description of the members, cf. `prepare_binop_pairs_between_dataframe_and_dataframe`.
    # This is a helper class to hold the results of this function.
    identifier: str
    lhs: SnowparkColumn
    lhs_datatype: Callable
    rhs: SnowparkColumn
    rhs_datatype: Callable


def prepare_binop_pairs_between_dataframe_and_dataframe(
    aligned_rhs_and_lhs: JoinOrAlignInternalFrameResult,
    combined_data_labels: list[Hashable],
    lhs_frame: InternalFrame,
    rhs_frame: InternalFrame,
) -> list[BinaryOperationPair]:
    """
    Returns a list of BinaryOperationPair which can be used to carry out a binary operation between two dataframes.
    Each BinaryOperationPair consists of the following:
    - identifier: an identifier that can be used within align_result to hold the result of a binary operation between two columns
    - lhs: a SnowparkColumn expression for the left operand
    - lhs_typer: a function to lazily determine the Snowpark datatype of `lhs`
    - rhs: a SnowparkColumn expression for the right operand
    - rhs_typer: a function to lazily determine the Snowpark datatype of `rhs`

    BinaryOperationPair will be returned in the order of `combined_data_labels`

    Args:
        aligned_rhs_and_lhs: the align result between other_frame and self_frame
        combined_data_labels: the combined data labels to be used for align result.
        rhs_frame: a frame representing the right side.
        lhs_frame: a frame representing the left side.

    Returns:
        List of BinaryOperationPair.
    """
    # construct list of pairs which label belongs to which quoted identifier
    type_map = aligned_rhs_and_lhs.result_frame.quoted_identifier_to_snowflake_type()
    left_right_pairs = []
    for label in combined_data_labels:
        left_identifier, right_identifier = None, None

        try:
            left_idx = lhs_frame.data_column_pandas_labels.index(label)
            left_quoted_identifier = lhs_frame.data_column_snowflake_quoted_identifiers[
                left_idx
            ]
            left_identifier = (
                aligned_rhs_and_lhs.result_column_mapper.left_quoted_identifiers_map[
                    left_quoted_identifier
                ]
            )
            left = col(left_identifier)
            # To avoid referencing always the last right_identifier in the loop, use functools.partial
            left_typer = functools.partial(
                lambda identifier: type_map[identifier], left_identifier
            )  # noqa: E731
        except ValueError:
            # lhs label not in list.
            left = pandas_lit(None)
            left_typer = lambda: infer_object_type(  # type: ignore[assignment] # noqa: E731
                None
            )

        try:
            right_idx = rhs_frame.data_column_pandas_labels.index(label)
            right_quoted_identifier = (
                rhs_frame.data_column_snowflake_quoted_identifiers[right_idx]
            )
            right_identifier = (
                aligned_rhs_and_lhs.result_column_mapper.right_quoted_identifiers_map[
                    right_quoted_identifier
                ]
            )
            right = col(right_identifier)
            # To avoid referencing always the last right_identifier in the loop, use functools.partial
            right_typer = functools.partial(
                lambda identifier: type_map[identifier], right_identifier
            )  # noqa: E731
        except ValueError:
            # rhs label not in list
            right = pandas_lit(None)
            right_typer = lambda: infer_object_type(None)  # type: ignore[assignment] # noqa: E731

        identifier_to_replace = left_identifier or right_identifier
        assert identifier_to_replace, "either identifier must be valid"

        # We return a lambda to determine the datatype of each operand lazily as this allows to defer
        # invoking a DESCRIBE query as much as possible.
        left_right_pairs.append(
            BinaryOperationPair(
                identifier=identifier_to_replace,
                lhs=left,
                lhs_datatype=left_typer,
                rhs=right,
                rhs_datatype=right_typer,
            )
        )
    return left_right_pairs



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

"""
File containing DataFrame APIs defined in the Modin API layer, but with different behavior in Snowpark
pandas, such as `DataFrame.memory_usage`.
"""

from typing import Any

import pandas as native_pd

from snowflake.snowpark.modin import pandas as pd  # noqa: F401
from snowflake.snowpark.modin.pandas import DataFrame, Series
from snowflake.snowpark.modin.pandas.api.extensions import register_dataframe_accessor
from snowflake.snowpark.modin.plugin._internal.telemetry import (
    snowpark_pandas_telemetry_method_decorator,
)
from snowflake.snowpark.modin.plugin.compiler.snowflake_query_compiler import (
    SnowflakeQueryCompiler,
)
from snowflake.snowpark.modin.plugin.utils.error_message import (
    dataframe_not_implemented,
)
from snowflake.snowpark.modin.utils import _inherit_docstrings


@_inherit_docstrings(native_pd.DataFrame.memory_usage, apilink="pandas.DataFrame")
@register_dataframe_accessor("memory_usage")
@snowpark_pandas_telemetry_method_decorator
def memory_usage(self, index: bool = True, deep: bool = False) -> Any:
    """
    Memory Usage (Dummy Information)

    The memory usage of a snowflake dataframe is not fully implemented.
    This method returns a series like the pandas dataframe to maintain
    compatibility with code which calls this for logging purposes, but
    the values are 0.

    Args:
        index: return dummy index memory usage
        deep: ignored

    Returns:
        Series with zeros for each index and column in the dataframe

    Examples:

        >>> df = pd.DataFrame({'Animal': ['Falcon', 'Falcon',
        ...                               'Parrot', 'Parrot'],
        ...                    'Max Speed': [380., 370., 24., 26.]})
        >>> df.memory_usage()
        Index        0
        Animal       0
        Max Speed    0
        dtype: int64

        >>> df.memory_usage(index=False)
        Animal       0
        Max Speed    0
        dtype: int64
    """
    # TODO: SNOW-1264697: push implementation down to query compiler
    columns = (["Index"] if index else []) + self._get_columns().array.tolist()
    return native_pd.Series([0] * len(columns), index=columns)


@_inherit_docstrings(native_pd.DataFrame.infer_objects, apilink="pandas.DataFrame")
@register_dataframe_accessor("infer_objects")
@snowpark_pandas_telemetry_method_decorator
@dataframe_not_implemented()
def infer_objects(
    self,
) -> SnowflakeQueryCompiler:  # pragma: no cover # noqa: RT01, D200
    """
    Attempt to infer better dtypes for object columns.
    """
    return self.__constructor__(query_compiler=self._query_compiler.infer_objects())


@_inherit_docstrings(native_pd.DataFrame.nunique, apilink="pandas.DataFrame")
@register_dataframe_accessor("nunique")
@snowpark_pandas_telemetry_method_decorator
def nunique(self, axis: int = 0, dropna: bool = True) -> Series:
    """
    Count number of distinct elements in specified axis.

    Return Series with number of distinct elements. Can ignore NaN
    values. Snowpark pandas API does not distinguish between NaN values and treats them all as the same.

    Parameters
    ----------
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to use. 0 or 'index' for row-wise, 1 or 'columns' for
        column-wise.
    dropna : bool, default True
        Don't include NaN in the counts.

    Returns
    -------
    Series

    Examples
    --------
    >>> import snowflake.snowpark.modin.pandas as pd
    >>> df = pd.DataFrame({'A': [4, 5, 6], 'B': [4, 1, 1]})
    >>> df.nunique()
    A    3
    B    2
    dtype: int8

    >>> df.nunique(axis=1)
    0    1
    1    2
    2    2
    dtype: int8

    >>> df = pd.DataFrame({'A': [None, pd.NA, None], 'B': [1, 2, 1]})
    >>> df.nunique()
    A    0
    B    2
    dtype: int8

    >>> df.nunique(dropna=False)
    A    1
    B    2
    dtype: int8

    """
    # TODO: SNOW-1264688: remove this override
    return super(DataFrame, self).nunique(axis=axis, dropna=dropna)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from collections.abc import Hashable
from typing import Union

import pandas as native_pd
from modin.core.dataframe.algebra.default2pandas import DataFrameDefault  # type: ignore

from snowflake.snowpark.functions import any_value, get
from snowflake.snowpark.modin.plugin._internal.frame import InternalFrame
from snowflake.snowpark.modin.plugin._internal.ordered_dataframe import (
    OrderedDataFrame,
    OrderingColumn,
)
from snowflake.snowpark.modin.plugin._internal.unpivot_utils import (
    UnpivotResultInfo,
    _prepare_unpivot_internal,
)
from snowflake.snowpark.modin.plugin._internal.utils import (
    INDEX_LABEL,
    LEVEL_LABEL,
    ROW_POSITION_COLUMN_LABEL,
    is_all_label_components_none,
    is_json_serializable_pandas_labels,
    pandas_lit,
    parse_object_construct_snowflake_quoted_identifier_and_extract_pandas_label,
    serialize_pandas_labels,
)

TRANSPOSE_INDEX = "TRANSPOSE_IDX"
# transpose value column used in unpivot
TRANSPOSE_VALUE_COLUMN = "TRANSPOSE_VAL"
# transpose name column used in unpivot
TRANSPOSE_NAME_COLUMN = "TRANSPOSE_COL_NAME"
# transpose json parsed object name
TRANSPOSE_OBJ_NAME_COLUMN = "TRANSPOSE_OBJ_NAME"


def transpose_empty_df(
    original_frame: InternalFrame,
) -> "SnowflakeQueryCompiler":  # type: ignore[name-defined] # noqa: F821
    from snowflake.snowpark.modin.plugin.compiler.snowflake_query_compiler import (
        SnowflakeQueryCompiler,
    )

    return SnowflakeQueryCompiler.from_pandas(
        native_pd.DataFrame(
            columns=original_frame.index_columns_index,
            index=original_frame.data_columns_index,
        )
    )


def prepare_and_unpivot_for_transpose(
    original_frame: InternalFrame,
    query_compiler: "SnowflakeQueryCompiler",  # type: ignore[name-defined] # noqa: F821
    is_single_row: bool = False,
) -> Union[UnpivotResultInfo, "SnowflakeQueryCompiler"]:  # type: ignore[name-defined] # noqa: F821

    # Check if the columns are all json serializable, if not, then go through fallback path.  The transpose approach
    # here requires json serializable labels because we use sql parse_json to split out row position and multi-level
    # index values as described below.
    #
    # TODO (SNOW-886400) Multi-level non-json serializable pandas label not handled.
    if not is_json_serializable_pandas_labels(original_frame.data_column_pandas_labels):
        return DataFrameDefault.register(native_pd.DataFrame.transpose)(query_compiler)

    # Ensure there is a row position since preserving order is important for unpivot and transpose.
    original_frame = original_frame.ensure_row_position_column()

    # Transpose is implemented with unpivot followed by pivot. However when the input dataframe is empty, there are two issues
    # 1) unpivot on empty table returns empty, which results in missing values in UNPIVOT_NAME_COLUMN
    # 2) pivot values can not be empty.
    # In order to overcome these, we add a dummy row to ordered_dataframe with row position value -1 to make sure
    # there is always atleast one row in the table, and drop the dummy column associated with row position -1 after pivot.
    ordered_dataframe = original_frame.ordered_dataframe
    row_position_snowflake_quoted_identifier = (
        original_frame.row_position_snowflake_quoted_identifier
    )
    if not is_single_row:
        quoted_identifiers = (
            ordered_dataframe.projected_column_snowflake_quoted_identifiers
        )
        new_columns = []
        for identifier in quoted_identifiers:
            if identifier == row_position_snowflake_quoted_identifier:
                new_columns.append((pandas_lit(-1)).as_(identifier))
            else:
                # We use any_value to select any value in the dummy column to make sure its dtypes are
                # the same as the column in the original dataframe. This helps avoid type incompatibility
                # issues in union_all.
                new_columns.append(any_value(identifier).as_(identifier))
        dummy_df = ordered_dataframe.agg(new_columns)
        ordered_dataframe = ordered_dataframe.union_all(dummy_df)

    return _prepare_unpivot_internal(
        original_frame=original_frame,
        ordered_dataframe=ordered_dataframe,
        is_single_row=is_single_row,
        index_column_name=TRANSPOSE_INDEX,
        value_column_name=TRANSPOSE_VALUE_COLUMN,
        variable_column_name=TRANSPOSE_NAME_COLUMN,
        object_column_name=TRANSPOSE_OBJ_NAME_COLUMN,
    )


def clean_up_transpose_result_index_and_labels(
    original_frame: InternalFrame,
    ordered_transposed_df: OrderedDataFrame,
    transpose_name_quoted_snowflake_identifier: str,
    transpose_object_name_quoted_snowflake_identifier: str,
) -> InternalFrame:
    """
    Creates an internal frame based on the original frame and the data transposed snowpark dataframe.  This
    cleans up and normalizes the labels and index values so they conform with expectations for pandas transpose.

    Example:
        If the original frame had:
            data column labels ('a', 'x'), ('a', 'y'), ('b', 'w'), ('b', 'z') and index column values (g, h, i)
        and transposed snowpark dataframe had:
            schema ('"TRANSPOSE_OBJ_NAME"',
                '"{""0"":""g"", ""row"":0}"', '"{""0"":""h"", ""row"":1}"', '"{""0"":""i"", ""row"":2}"')
            and values for TRANSPOSE_OBJ_NAME: [0, ["a", "x"]], [1, ["a", "y"]], [2, ["b", "w"]], [3, ["b", "z"]]
        then the dataframe index is split into multi-columns and labels are cleaned up.

        The resulting frame would have (transposed indexes):
            data column labels: (g, h, i) and index column values ('a', 'x'), ('a', 'y'), ('b', 'w'), ('b', 'z')
        and normalized snowpark dataframe:
            schema ('"row_position"', '"level"', '"level_1"', '"g"', '"h"' ,'"i"')
            and values (0, a, x), (1, a, y), (2, b, w), (3, b, z) for values __row_position, level, level_1

    Args:
        original_frame: The original InternalFrame for the transpose
        ordered_transposed_df: The transposed ordered dataframe
        transpose_name_quoted_snowflake_identifier: variable name identifier from the unpivot
        transpose_object_name_quoted_snowflake_identifier: values from the unpivot

    Returns:
        The transposed InternalFrame.
    """
    # The remaining columns are the resulting output columns of the transpose, except for the TRANSPOSE_NAME_COLUMN
    # which becomes the new index of the resulting table.
    data_column_snowflake_quoted_identifiers = (
        ordered_transposed_df.projected_column_snowflake_quoted_identifiers
    )
    data_column_snowflake_quoted_identifiers.remove(
        transpose_name_quoted_snowflake_identifier
    )
    data_column_snowflake_quoted_identifiers.remove(
        transpose_object_name_quoted_snowflake_identifier
    )
    data_column_object_identifier_pairs = [
        (
            parse_object_construct_snowflake_quoted_identifier_and_extract_pandas_label(
                snowflake_quoted_identifier,
                len(original_frame.index_column_pandas_labels),
            ),
            snowflake_quoted_identifier,
        )
        for snowflake_quoted_identifier in data_column_snowflake_quoted_identifiers
    ]

    # Extract the position information that was previously serialized into the column names, then sort and
    # re-organize the column names to maintain the original ordering from the pre-transpose rows.
    data_column_object_identifier_pairs.sort(
        key=lambda obj_ident: obj_ident[0][1]["row"]
    )

    # Drop the identifiers associated with dummy column row:-1 generated from the dummy row in transpose.
    if len(data_column_object_identifier_pairs) > 0:
        if data_column_object_identifier_pairs[0][0][1]["row"] == -1:
            data_column_object_identifier_pairs.remove(
                data_column_object_identifier_pairs[0]
            )

    # If it's a single level, we store the label, otherwise we store tuple for each level.
    new_data_column_pandas_labels = [
        data_column_object_identifier[0]
        for data_column_object_identifier, _ in data_column_object_identifier_pairs
    ]

    new_data_column_snowflake_quoted_identifiers = [
        snowflake_quoted_identifier
        for _, snowflake_quoted_identifier in data_column_object_identifier_pairs
    ]

    # We need to split out the TRANSPOSE_OBJ_NAME_COLUMN with two cases:
    #
    # If it is a single index, the format will be [1, "employed"] and result in new columns with values:
    #       (row_position, 1), ("__level__", "employed")
    #
    # If it is a multi-index, the format will be [1, ["status", "employed"]] and result in new columns with values:
    #       (row_position, 1), ("__level_1__", "status"), ("__level_2__", "employed")
    new_index_column_pandas_labels: list[Hashable] = []
    new_index_column_snowflake_quoted_identifiers: list[str] = []
    for i, pandas_label in enumerate(original_frame.data_column_pandas_index_names):
        if is_all_label_components_none(pandas_label):
            index_label = LEVEL_LABEL
            if i >= 1:
                index_label += f"_{i}"
        else:
            index_label = pandas_label

        snowflake_quoted_identifier = (
            ordered_transposed_df.generate_snowflake_quoted_identifiers(
                pandas_labels=serialize_pandas_labels([index_label]),
                excluded=new_data_column_snowflake_quoted_identifiers
                + new_index_column_snowflake_quoted_identifiers,
            )[0]
        )

        new_index_column_pandas_labels.append(pandas_label)
        new_index_column_snowflake_quoted_identifiers.append(
            snowflake_quoted_identifier
        )

    # Extract the new row position and pandas label object from column
    # transpose_object_name_quoted_snowflake_identifier, which is an array column
    # with value [row_position, label object] like [0, "score"]. The label object
    # for multi-index can look like {"0": "A", "1": "B"} for panda label ("A", "B").

    # Generate the snowflake quoted identifier for extracted row position and pandas
    # label object columns.
    row_position_and_index_snowflake_quoted_identifier = (
        ordered_transposed_df.generate_snowflake_quoted_identifiers(
            pandas_labels=[ROW_POSITION_COLUMN_LABEL, INDEX_LABEL],
            excluded=new_data_column_snowflake_quoted_identifiers
            + new_index_column_snowflake_quoted_identifiers,
        )
    )
    pivot_with_index_select_list = [
        get(transpose_object_name_quoted_snowflake_identifier, i).as_(
            snowflake_quoted_identifier
        )
        for i, snowflake_quoted_identifier in enumerate(
            row_position_and_index_snowflake_quoted_identifier
        )
    ] + new_data_column_snowflake_quoted_identifiers

    ordered_transposed_df = ordered_transposed_df.select(pivot_with_index_select_list)

    row_position_snowflake_quoted_identifier = (
        row_position_and_index_snowflake_quoted_identifier[0]
    )
    index_snowflake_quoted_identifier = (
        row_position_and_index_snowflake_quoted_identifier[1]
    )
    # Handle the multi-index case by further parsing out each level to a separate level_# columns.
    if len(new_index_column_snowflake_quoted_identifiers) > 1:
        pivot_with_multi_index_select_list = (
            [row_position_snowflake_quoted_identifier]
            + [
                get(index_snowflake_quoted_identifier, i).as_(
                    snowflake_quoted_identifier
                )
                for i, snowflake_quoted_identifier in enumerate(
                    new_index_column_snowflake_quoted_identifiers
                )
            ]
            + new_data_column_snowflake_quoted_identifiers
        )

        ordered_transposed_df = ordered_transposed_df.select(
            pivot_with_multi_index_select_list
        )
    else:
        # If it is a single level then no more extraction is needed after separating the row position and index.
        new_index_column_snowflake_quoted_identifiers = [
            index_snowflake_quoted_identifier
        ]

    # Create new internal frame with resulting ordering column and transposed index values.
    ordered_transposed_df = ordered_transposed_df.sort(
        OrderingColumn(row_position_snowflake_quoted_identifier)
    )

    new_internal_frame = InternalFrame.create(
        ordered_dataframe=ordered_transposed_df,
        data_column_pandas_labels=new_data_column_pandas_labels,
        data_column_pandas_index_names=original_frame.index_column_pandas_labels,
        data_column_snowflake_quoted_identifiers=new_data_column_snowflake_quoted_identifiers,
        index_column_pandas_labels=new_index_column_pandas_labels,
        index_column_snowflake_quoted_identifiers=new_index_column_snowflake_quoted_identifiers,
    )

    # Rename the data column snowflake quoted identifiers to be closer to pandas labels, normalizing names
    # will remove information like row position that may have temporarily been included in column names to track
    # during earlier steps.
    new_internal_frame = (
        new_internal_frame.normalize_snowflake_quoted_identifiers_with_pandas_label()
    )

    return new_internal_frame



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

# Code in this file may constitute partial or total reimplementation, or modification of
# existing code originally distributed by the Modin project, under the Apache License,
# Version 2.0.

"""Implement GroupBy public API as pandas does."""

from collections.abc import Hashable
from typing import Any, Callable, Literal, Optional, Sequence, Union

import numpy as np  # noqa: F401
import numpy.typing as npt
import pandas
import pandas.core.groupby
from pandas._libs.lib import NoDefault, no_default
from pandas._typing import AggFuncType, Axis, IndexLabel
from pandas.core.dtypes.common import is_dict_like, is_list_like, is_numeric_dtype
from pandas.errors import SpecificationError
from pandas.io.formats.printing import PrettyDict
from pandas.util._validators import validate_bool_kwarg

# the following import is used in doctests
from snowflake.snowpark.modin import pandas as pd  # noqa: F401

# Snowpark pandas API version
from snowflake.snowpark.modin.pandas.series import Series
from snowflake.snowpark.modin.pandas.utils import (
    extract_validate_and_try_convert_named_aggs_from_kwargs,
    raise_if_native_pandas_objects,
    validate_and_try_convert_agg_func_arg_func_to_str,
)
from snowflake.snowpark.modin.plugin._internal.apply_utils import (
    create_groupby_transform_func,
)
from snowflake.snowpark.modin.plugin._internal.telemetry import TelemetryMeta
from snowflake.snowpark.modin.plugin.compiler.snowflake_query_compiler import (
    SnowflakeQueryCompiler,
)
from snowflake.snowpark.modin.plugin.utils.error_message import ErrorMessage
from snowflake.snowpark.modin.plugin.utils.warning_message import WarningMessage
from snowflake.snowpark.modin.utils import (
    MODIN_UNNAMED_SERIES_LABEL,
    _inherit_docstrings,
    doc_replace_dataframe_with_link,
    hashable,
    validate_int_kwarg,
)

_DEFAULT_BEHAVIOUR = {
    "__class__",
    "__getitem__",
    "__init__",
    "__iter__",
    "_as_index",
    "_axis",
    "_by",
    "_check_index_name",
    "_columns",
    "_df",
    "_groups_cache",
    "_idx_name",
    "_index",
    "_indices_cache",
    "_internal_by",
    "_internal_by_cache",
    "_iter",
    "_kwargs",
    "_level",
    "_pandas_class",
    "_query_compiler",
    "_sort",
    "_wrap_aggregation",
}


@_inherit_docstrings(
    pandas.core.groupby.DataFrameGroupBy, modify_doc=doc_replace_dataframe_with_link
)
class DataFrameGroupBy(metaclass=TelemetryMeta):
    _pandas_class = pandas.core.groupby.DataFrameGroupBy

    def __init__(
        self,
        df,
        by,
        axis,
        level,
        as_index,
        sort,
        group_keys,
        idx_name,
        **kwargs,
    ) -> None:
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        self._axis = axis
        self._idx_name = idx_name
        self._df = df
        self._df._query_compiler.validate_groupby(by, axis, level)
        self._query_compiler = self._df._query_compiler
        self._columns = self._query_compiler.columns
        self._by = by
        self._level = level
        self._kwargs = {
            "level": level,
            "sort": sort,
            "as_index": as_index,
            "group_keys": group_keys,
        }
        self._kwargs.update(kwargs)

    def __getattr__(self, key):
        """
        Alter regular attribute access, looks up the name in the columns.

        Parameters
        ----------
        key : str
            Attribute name.

        Returns
        -------
        The value of the attribute.
        """
        try:
            return object.__getattribute__(self, key)
        except AttributeError as err:
            if key in self._columns:
                return self.__getitem__(key)
            raise err

    @property
    def ngroups(self):
        return self._query_compiler.groupby_ngroups(
            by=self._by,
            axis=self._axis,
            groupby_kwargs=self._kwargs,
        )

    def skew(self, *args, **kwargs):
        ErrorMessage.method_not_implemented_error(name="skew", class_="GroupBy")

    def ffill(self, limit=None):
        ErrorMessage.method_not_implemented_error(name="ffill", class_="GroupBy")

    def sem(self, ddof=1):
        ErrorMessage.method_not_implemented_error(name="sem", class_="GroupBy")

    def value_counts(
        self,
        subset=None,
        normalize: bool = False,
        sort: bool = True,
        ascending: bool = False,
        dropna: bool = True,
    ):
        ErrorMessage.method_not_implemented_error(name="value_counts", class_="GroupBy")

    def mean(
        self,
        numeric_only: bool = False,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_mean", engine, engine_kwargs
        )
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=numeric_only,
            agg_func="mean",
            agg_kwargs=dict(numeric_only=numeric_only),
        )

    def any(self, skipna=True):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="any", class_="GroupBy")

    @property
    def plot(self):  # pragma: no cover
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="plot", class_="GroupBy")

    def ohlc(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="ohlc", class_="GroupBy")

    def __bytes__(self):
        """
        Convert DataFrameGroupBy object into a python2-style byte string.

        Returns
        -------
        bytearray
            Byte array representation of `self`.

        Notes
        -----
        Deprecated and removed in pandas and will be likely removed in Modin.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="__bytes__", class_="GroupBy")

    _groups_cache = no_default

    # TODO: since python 3.9:
    # @cached_property
    @property
    def groups(self) -> PrettyDict[Hashable, pd.Index]:
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._query_compiler.groupby_groups(
            self._by,
            self._axis,
            groupby_kwargs={
                # groupby.groups always treat as_index as True. this seems to be
                # intended behavior: https://github.com/pandas-dev/pandas/issues/56965
                k: True if k == "as_index" else v
                for k, v in self._kwargs.items()
            },
        )

    def min(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
    ):
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_min", engine, engine_kwargs
        )
        validate_int_kwarg(min_count, "min_count", float_allowed=False)
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=numeric_only,
            agg_func="min",
            agg_kwargs=dict(min_count=min_count, numeric_only=numeric_only),
        )

    def idxmax(
        self, axis: Axis = no_default, skipna: bool = True, numeric_only: bool = False
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        axis_number = self._df._get_axis_number(axis)
        if axis_number == 1:
            # Performing idxmax is deprecated and will be removed in a future pandas version.
            raise NotImplementedError(
                "DataFrameGroupBy.idxmax with axis=1 is deprecated and will be removed in a "
                "future version. Operate on the un-grouped DataFrame instead."
            )
        else:
            # When performing idxmax/idxmin on axis=0, it can be done column-wise.
            result = self._wrap_aggregation(
                qc_method=type(self._query_compiler).groupby_agg,
                numeric_only=numeric_only,
                how="axis_wise",
                agg_func="idxmax",
                # axis is also specified here since the axis used with idxmax/idxmin is different from
                # the groupby axis.
                agg_kwargs=dict(skipna=skipna, axis=0),
            )
        return result

    def idxmin(
        self, axis: Axis = no_default, skipna: bool = True, numeric_only: bool = False
    ) -> Series:
        axis_number = self._df._get_axis_number(axis)
        if axis_number == 1:
            # Performing idxmin is deprecated and will be removed in a future pandas version.
            raise NotImplementedError(
                "DataFrameGroupBy.idxmin with axis=1 is deprecated and will be removed in a "
                "future version. Operate on the un-grouped DataFrame instead."
            )
        else:
            # When performing idxmax/idxmin on axis=0, it can be done column-wise.
            result = self._wrap_aggregation(
                qc_method=type(self._query_compiler).groupby_agg,
                numeric_only=numeric_only,
                how="axis_wise",
                agg_func="idxmin",
                # axis is also specified here since the axis used with idxmax/idxmin is different from
                # the groupby axis.
                agg_kwargs=dict(skipna=skipna, axis=0),
            )
        return result

    @property
    def ndim(self):
        """
        Return 2.

        Returns
        -------
        int
            Returns 2.

        Notes
        -----
        Deprecated and removed in pandas and will be likely removed in Modin.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return 2  # ndim is always 2 for DataFrames

    def shift(
        self,
        periods: Union[int, Sequence[int]] = 1,
        freq: int = None,
        axis: Axis = 0,
        fill_value: Any = None,
        suffix: Optional[str] = None,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        if isinstance(periods, Sequence):
            ErrorMessage.not_implemented(
                "Snowpark pandas GroupBy.shift does not yet support `periods` that are sequences. Only int `periods` are supported."
            )
        if suffix is not None:
            ErrorMessage.not_implemented(
                "Snowpark pandas GroupBy.shift does not yet support the `suffix` parameter"
            )
        if not isinstance(periods, int):
            raise TypeError(
                f"Periods must be integer, but {periods} is {type(periods)}."
            )
        qc = self._query_compiler.groupby_shift(
            self._by, self._axis, self._level, periods, freq, fill_value, self.ndim == 1
        )
        return (
            pd.Series(query_compiler=qc)
            if self.ndim == 1
            else pd.DataFrame(query_compiler=qc)
        )

    def nth(self, n, dropna=None):
        ErrorMessage.method_not_implemented_error(name="nth", class_="GroupBy")

    def cumsum(self, axis: Axis = 0, *args, **kwargs):
        qc = self._query_compiler.groupby_cumsum(self._by, self._axis, self._kwargs)
        return (
            pd.Series(query_compiler=qc)
            if self.ndim == 1
            else pd.DataFrame(query_compiler=qc)
        )

    @property
    def indices(self) -> dict[Hashable, npt.NDArray[np.intp]]:
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._query_compiler.groupby_indices(
            self._by,
            self._axis,
            groupby_kwargs={
                # groupby.indices always treat as_index as True. this seems to be
                # intended behavior: https://github.com/pandas-dev/pandas/issues/56965
                k: True if k == "as_index" else v
                for k, v in self._kwargs.items()
            },
        )

    @_inherit_docstrings(pandas.core.groupby.DataFrameGroupBy.pct_change)
    def pct_change(self, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="pct_change", class_="GroupBy")

    def filter(self, func, dropna=True, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="filter", class_="GroupBy")

    def cummax(self, axis: Axis = 0, numeric_only: bool = False, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        qc = self._query_compiler.groupby_cummax(
            self._by, self._axis, numeric_only, self._kwargs
        )
        return (
            pd.Series(query_compiler=qc)
            if self.ndim == 1
            else pd.DataFrame(query_compiler=qc)
        )

    def apply(self, func, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        # TODO: SNOW-1244717: Explore whether window function are performant and can be used
        #       whenever `func` is an aggregation function.
        if not callable(func):
            raise NotImplementedError("No support for non-callable `func`")
        dataframe_result = pd.DataFrame(
            query_compiler=self._query_compiler.groupby_apply(
                self._by,
                agg_func=func,
                axis=self._axis,
                groupby_kwargs=self._kwargs,
                agg_args=args,
                agg_kwargs=kwargs,
                series_groupby=False,
            )
        )
        if dataframe_result.columns.equals(pandas.Index([MODIN_UNNAMED_SERIES_LABEL])):
            return dataframe_result.squeeze(axis=1)
        return dataframe_result

    @property
    def dtypes(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="dtypes", class_="GroupBy")

    def first(self, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="first", class_="GroupBy")

    _internal_by_cache = no_default

    # TODO: since python 3.9:
    # @cached_property
    @property
    def _internal_by(self):
        """
        Get only those components of 'by' that are column labels of the source frame.

        Returns
        -------
        tuple of labels
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        if self._internal_by_cache is not no_default:
            return self._internal_by_cache

        by_list = self._by if is_list_like(self._by) else [self._by]

        internal_by = tuple(
            by for by in by_list if hashable(by) and by in self._columns
        )

        self._internal_by_cache = internal_by
        return internal_by

    def __getitem__(self, key):
        """
        Implement indexing operation on a DataFrameGroupBy object.

        Parameters
        ----------
        key : list or str
            Names of columns to use as subset of original object.

        Returns
        -------
        DataFrameGroupBy or SeriesGroupBy
            Result of indexing operation.

        Raises
        ------
        NotImplementedError
            Column lookups on GroupBy with arbitrary Series in by is not yet supported.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        if self._axis == 1:
            raise ValueError("Cannot subset columns when using axis=1")

        # These parameters are common for building the resulted Series or DataFrame groupby object
        kwargs = {
            **self._kwargs.copy(),
            "by": self._by,
            "axis": self._axis,
            "idx_name": self._idx_name,
        }
        # The rules of type deduction for the resulted object is the following:
        #   1. If `key` is a list-like or `as_index is False`, then the resulted object is a DataFrameGroupBy
        #   2. Otherwise, the resulted object is SeriesGroupBy
        #   3. Result type does not depend on the `by` origin
        # Examples:
        #   - drop: any, as_index: any, __getitem__(key: list_like) -> DataFrameGroupBy
        #   - drop: any, as_index: False, __getitem__(key: any) -> DataFrameGroupBy
        #   - drop: any, as_index: True, __getitem__(key: label) -> SeriesGroupBy
        if is_list_like(key):
            make_dataframe = True
        else:
            if self._as_index:
                make_dataframe = False
            else:
                make_dataframe = True
            key = [key]

        column_index = self._df.columns
        # validate that all keys are labels belong to the data column of the df
        for label in key:
            if not (label in column_index):
                raise KeyError(f"Columns not found: '{label}'")

        # internal_by records all label in by that belongs to the data columns
        internal_by = frozenset(self._internal_by)
        if len(internal_by.intersection(key)) != 0:
            message = (
                "Data column selection with overlap of 'by' columns is not yet supported, "
                "please duplicate the overlapped by columns and rename it to a different name"
            )
            ErrorMessage.not_implemented(message=message)

        # select the union of the internal bys and select keys. Here we find all integer
        # positions for all the selected columns, and then call iloc to select all columns.
        # This is because loc currently doesn't support select with multiindex, once iloc and
        # dataframe getitem is supported, this can be replaced with df[list(internal_by) + list(key)]
        # TODO (SNOW-896342): update self._df.iloc[:, ilocs_list] to use df[list(internal_by) + list(key)]
        #           once dataframe getitem is supported.
        _, by_ilocs = column_index._get_indexer_strict(list(internal_by), "columns")
        _, key_ilocs = column_index._get_indexer_strict(list(key), "columns")
        ilocs_list = list(by_ilocs) + list(key_ilocs)

        if len(key_ilocs) > 1:
            make_dataframe = True

        if make_dataframe:
            return DataFrameGroupBy(
                self._df.iloc[:, ilocs_list],
                **kwargs,
            )
        else:
            return SeriesGroupBy(
                self._df.iloc[:, ilocs_list],
                **kwargs,
            )

    def cummin(self, axis: Axis = 0, numeric_only: bool = False, *args, **kwargs):
        qc = self._query_compiler.groupby_cummin(
            self._by, self._axis, numeric_only, self._kwargs
        )
        return (
            pd.Series(query_compiler=qc)
            if self.ndim == 1
            else pd.DataFrame(query_compiler=qc)
        )

    def bfill(self, limit=None):
        ErrorMessage.method_not_implemented_error(name="bfill", class_="GroupBy")

    def prod(self, numeric_only=False, min_count=0):
        ErrorMessage.method_not_implemented_error(name="prod", class_="GroupBy")

    def std(
        self,
        ddof: int = 1,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
        numeric_only: bool = False,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_std", engine, engine_kwargs
        )
        return self._wrap_aggregation(
            qc_method=SnowflakeQueryCompiler.groupby_agg,
            numeric_only=numeric_only,
            agg_func="std",
            agg_kwargs=dict(ddof=ddof, numeric_only=numeric_only),
        )

    def aggregate(
        self,
        func: Optional[AggFuncType] = None,
        *args: Any,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
        **kwargs: Any,
    ):
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_aggregate", engine, engine_kwargs
        )
        if self._axis != 0 and (is_dict_like(func) or is_list_like(func)):
            # This is the same as pandas for func that is a list or dict
            ErrorMessage.not_implemented(
                "axis other than 0 is not supported"
            )  # pragma: no cover
        if func is None:
            func = extract_validate_and_try_convert_named_aggs_from_kwargs(
                obj=self, allow_duplication=True, axis=self._axis, **kwargs
            )
        else:
            func = validate_and_try_convert_agg_func_arg_func_to_str(
                agg_func=func,
                obj=self,
                allow_duplication=True,
                axis=self._axis,
            )

        if isinstance(func, str):
            # Using "getattr" here masks possible AttributeError which we throw
            # in __getattr__, so we should call __getattr__ directly instead.
            agg_func = self.__getattr__(func)
            if callable(agg_func):
                return agg_func(*args, **kwargs)

        # when the aggregation function passed in is list like always return a Dataframe regardless
        # it is SeriesGroupBy or DataFrameGroupBy
        is_result_dataframe = (self.ndim == 2) or is_list_like(func)
        result = self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=False,
            agg_func=func,
            agg_args=args,
            agg_kwargs=kwargs,
            how="axis_wise",
            is_result_dataframe=is_result_dataframe,
        )
        return result

    agg = aggregate

    def last(self, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="last", class_="GroupBy")

    def rank(
        self,
        method: str = "average",
        ascending: bool = True,
        na_option: str = "keep",
        pct: bool = False,
        *args,
        **kwargs,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        query_compiler = self._query_compiler.groupby_rank(
            by=self._by,
            axis=self._axis,
            method=method,
            na_option=na_option,
            ascending=ascending,
            pct=pct,
            groupby_kwargs=self._kwargs,
            agg_args=args,
            agg_kwargs=kwargs,
        )
        if self.ndim == 1:
            result = pd.Series(query_compiler=query_compiler)
        else:
            result = pd.DataFrame(query_compiler=query_compiler)
        return result

    @property
    def corrwith(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="corrwith", class_="GroupBy")

    def max(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_max", engine, engine_kwargs
        )
        validate_int_kwarg(min_count, "min_count", float_allowed=False)
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=numeric_only,
            agg_func="max",
            agg_kwargs=dict(min_count=min_count, numeric_only=numeric_only),
        )

    def var(
        self,
        ddof: int = 1,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
        numeric_only: bool = False,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_var", engine, engine_kwargs
        )

        return self._wrap_aggregation(
            qc_method=SnowflakeQueryCompiler.groupby_agg,
            numeric_only=numeric_only,
            agg_func="var",
            agg_kwargs=dict(ddof=ddof, numeric_only=numeric_only),
        )

    def get_group(self, name, obj=None):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="get_group", class_="GroupBy")

    def __len__(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="__len__", class_="GroupBy")

    def all(self, skipna=True):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="all", class_="GroupBy")

    def size(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="size", class_="GroupBy")

    def sum(
        self,
        numeric_only: bool = False,
        min_count: int = 0,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_sum", engine, engine_kwargs
        )
        validate_int_kwarg(min_count, "min_count", float_allowed=False)
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=numeric_only,
            agg_func="sum",
            agg_kwargs=dict(min_count=min_count, numeric_only=numeric_only),
        )

    def describe(self, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="describe", class_="GroupBy")

    def boxplot(
        self,
        grouped,
        subplots=True,
        column=None,
        fontsize=None,
        rot=0,
        grid=True,
        ax=None,
        figsize=None,
        layout=None,
        **kwargs,
    ):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="boxplot", class_="GroupBy")

    def ngroup(self, ascending=True):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="ngroup", class_="GroupBy")

    def nunique(self, dropna=True):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_nunique,
            agg_func="nunique",
            agg_kwargs=dict(dropna=dropna),
        )

    def resample(self, rule, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="resample", class_="GroupBy")

    def sample(self, n=None, frac=None, replace=False, weights=None, random_state=None):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="sample", class_="GroupBy")

    def median(self, numeric_only: bool = False):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=numeric_only,
            agg_func="median",
            agg_kwargs=dict(numeric_only=numeric_only),
        )

    def head(self, n=5):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        # Ensure that n is an integer value.
        if not isinstance(n, int):
            raise TypeError("n must be an integer value.")

        # Only the groupby parameter "dropna" affects the output of head. None of the other groupby
        # parameters: as_index, sort, and group_keys, affect head.
        # Values needed for the helper functions.
        agg_kwargs = {
            "n": n,
            "level": self._level,
            "dropna": self._kwargs.get("dropna", True),
        }

        result = self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            agg_func="head",
            agg_kwargs=agg_kwargs,
        )
        return pd.DataFrame(result)

    def cumprod(self, axis=0, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="cumprod", class_="GroupBy")

    def __iter__(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._iter.__iter__()

    def cov(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="cov", class_="GroupBy")

    def transform(
        self,
        func: Union[str, Callable],
        *args: Any,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
        **kwargs: Any,
    ) -> "pd.DataFrame":
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        WarningMessage.warning_if_engine_args_is_set(
            "groupby_transform", engine, engine_kwargs
        )

        # The resultant DataFrame from `transform` has an index that always matches the original
        # DataFrame's index.
        # Create a new groupby object so that we can tune parameters to ensure that `apply`
        # returns a DataFrame with the required index (same as original DataFrame).
        #
        # Unlike `transform`, groupby parameters affect the result of `apply`.
        # - `group_keys` controls whether the grouped column(s) are included in the index.
        #   `group_keys` needs to be False to ensure that the resultant DataFrame has the
        #   original DataFrame's index.
        #
        # - `dropna` controls whether the NA values should be included as a group/be present
        #    in the group keys. `transform` always includes the NA values, therefore `dropna`
        #    needs to be False to ensure that all NA values are included.
        #
        # - `sort` controls whether the group keys are sorted.
        #
        # - `as_index` controls whether the groupby object has group labels as the index.
        by = self._by
        level = self._level
        groupby_obj = self._df.groupby(
            by=by,  # either by or levels can be specified at a time
            level=level,
            as_index=self._as_index,
            group_keys=False,
            dropna=False,
            sort=self._sort,
        )

        # Apply the transform function to each group.
        res = groupby_obj.apply(
            create_groupby_transform_func(func, by, level, *args, **kwargs)
        )

        dropna = self._kwargs.get("dropna", True)
        if dropna is True:
            # - To avoid dropping any NA values, `dropna` is set to False in both the groupby
            #   object created above and the groupby object created in `create_groupby_transform_func`.
            #
            # - If dropna is set to True in the groupby object, the output from this code (so far)
            #   and the expected native pandas result differs.
            #
            # - In the Snowpark pandas code, all rows grouped under NA keys calculate the result with
            #   the given `func`, thus resulting in non-NA values.
            #
            # - In the native pandas version, all rows grouped under NA keys take up
            #   "NaN" values in all columns.
            #
            # Therefore, we need to convert the rows grouped under NA keys to have NaN values in
            # all columns.
            na_col_data = self._df[by].isna()
            condition = (
                na_col_data.any(axis=1)
                if isinstance(na_col_data, pd.DataFrame)
                else na_col_data
            )
            res.loc[condition, :] = np.nan

        return res

    def corr(self, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="corr", class_="GroupBy")

    def fillna(self, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="fillna", class_="GroupBy")

    def count(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        result = self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            numeric_only=False,
            agg_func="count",
        )
        return result

    def pipe(self, func, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="pipe", class_="GroupBy")

    def cumcount(self, ascending: bool = True):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        query_compiler = self._query_compiler.groupby_cumcount(
            self._by, self._axis, self._kwargs, ascending
        )
        return pd.Series(query_compiler=query_compiler)

    def tail(self, n=5):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        # Ensure that n is an integer value.
        if not isinstance(n, int):
            raise TypeError("n must be an integer value.")

        # Only the groupby parameter "dropna" affects the output of tail. None of the other groupby
        # parameters: as_index, sort, and group_keys, affect tail.
        # Values needed for the helper functions.
        agg_kwargs = {
            "n": n,
            "level": self._level,
            "dropna": self._kwargs.get("dropna", True),
        }

        result = self._wrap_aggregation(
            qc_method=type(self._query_compiler).groupby_agg,
            agg_func="tail",
            agg_kwargs=agg_kwargs,
        )
        return pd.DataFrame(result)

    # expanding and rolling are unique cases and need to likely be handled
    # separately. They do not appear to be commonly used.
    def expanding(self, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="expanding", class_="GroupBy")

    def rolling(self, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="rolling", class_="GroupBy")

    def hist(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="hist", class_="GroupBy")

    def quantile(self, q=0.5, interpolation="linear"):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._wrap_aggregation(
            type(self._query_compiler).groupby_agg,
            numeric_only=False,
            agg_func="quantile",
            agg_kwargs=dict(q=q, interpolation=interpolation),
        )

    def diff(self):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="diff", class_="GroupBy")

    def take(self, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="take", class_="GroupBy")

    @property
    def _index(self):
        """
        Get index value.

        Returns
        -------
        pandas.Index
            Index value.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._query_compiler.index

    @property
    def _sort(self):
        """
        Get sort parameter value.

        Returns
        -------
        bool
            Value of sort parameter used to create DataFrameGroupBy object.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._kwargs.get("sort")

    @property
    def _as_index(self):
        """
        Get as_index parameter value.

        Returns
        -------
        bool
            Value of as_index parameter used to create DataFrameGroupBy object.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        return self._kwargs.get("as_index")

    @property
    def _iter(self):
        """
        Construct a tuple of (group_id, DataFrame) tuples to allow iteration over groups.

        Returns
        -------
        generator
            Generator expression of GroupBy object broken down into tuples for iteration.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        ErrorMessage.method_not_implemented_error(name="_iter", class_="GroupBy")

    def _wrap_aggregation(
        self,
        qc_method: Callable,
        numeric_only: bool = False,
        agg_args: list[Any] = None,
        agg_kwargs: dict[str, Any] = None,
        is_result_dataframe: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Perform common metadata transformations and apply groupby functions.

        Parameters
        ----------
        qc_method : callable
            The query compiler method to call.
        numeric_only : bool, default: False
            Specifies whether to aggregate non numeric columns:
                - True: include only numeric columns (including categories that holds a numeric dtype)
                - False: include all columns
        agg_args : list-like, optional
            Positional arguments to pass to the aggregation function.
        agg_kwargs : dict-like, optional
            Keyword arguments to pass to the aggregation function.
        is_result_dataframe: bool optional
            whether the result of aggregation is a dataframe or series. If None, is_result_dataframe will be
            False for SeriesGroupBy, and True for DataFrameGroupBy.
        **kwargs : dict
            Keyword arguments to pass to the specified query compiler's method.

        Returns
        -------
        DataFrame or Series
            Returns the same type as `self._df`.
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        numeric_only = validate_bool_kwarg(
            numeric_only, "numeric_only", none_allowed=True
        )

        agg_args = tuple() if agg_args is None else agg_args
        agg_kwargs = dict() if agg_kwargs is None else agg_kwargs

        is_series_groupby = self.ndim == 1
        if is_series_groupby:
            # when ndim is 1, it is SeriesGroupBy. SeriesGroupBy does not implement numeric_only
            # parameter even if it accepts the parameter, and the aggregation is handled the
            # same as numeric_only is False.
            if numeric_only and not is_numeric_dtype(self._query_compiler.dtypes[0]):
                # pandas throws an NotImplementedError when the numeric_only is True, but the
                # series dtype is not numeric
                ErrorMessage.not_implemented(
                    "SeriesGroupBy does not implement numeric_only"
                )
            numeric_only = False

        if is_result_dataframe is None:
            is_result_dataframe = not is_series_groupby
        result_type = pd.DataFrame if is_result_dataframe else pd.Series
        result = result_type(
            query_compiler=qc_method(
                self._query_compiler,
                by=self._by,
                axis=self._axis,
                groupby_kwargs=self._kwargs,
                agg_args=agg_args,
                agg_kwargs=agg_kwargs,
                numeric_only=numeric_only,
                is_series_groupby=is_series_groupby,
                **kwargs,
            )
        )
        return result

    def _check_index_name(self, result):
        """
        Check the result of groupby aggregation on the need of resetting index name.

        Parameters
        ----------
        result : DataFrame
            Group by aggregation result.

        Returns
        -------
        DataFrame
        """
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.DataFrameGroupBy functions
        if self._by is not None:
            # pandas does not name the index for this case
            result._query_compiler.set_index_name(None)
        return result


@_inherit_docstrings(
    pandas.core.groupby.SeriesGroupBy, modify_doc=doc_replace_dataframe_with_link
)
class SeriesGroupBy(DataFrameGroupBy):
    _pandas_class = pandas.core.groupby.SeriesGroupBy

    @property
    def ndim(self):
        """
        Return 1.

        Returns
        -------
        int
            Returns 1.

        Notes
        -----
        Deprecated and removed in pandas and will be likely removed in Modin.
        """
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        return 1  # ndim is always 1 for Series

    @property
    def _iter(self):
        """
        Construct a tuple of (group_id, Series) tuples to allow iteration over groups.

        Returns
        -------
        generator
            Generator expression of GroupBy object broken down into tuples for iteration.
        """
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(name="_iter", class_="GroupBy")

    @property
    def is_monotonic_decreasing(self):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(
            name="is_monotonic_decreasing", class_="GroupBy"
        )

    @property
    def is_monotonic_increasing(self):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(
            name="is_monotonic_increasing", class_="GroupBy"
        )

    def aggregate(
        self,
        func: Optional[AggFuncType] = None,
        *args: Any,
        engine: Optional[Literal["cython", "numba"]] = None,
        engine_kwargs: Optional[dict[str, bool]] = None,
        **kwargs: Any,
    ):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        if is_dict_like(func):
            raise SpecificationError(
                "Value for func argument in dict format is not allowed for SeriesGroupBy."
            )

        return super().aggregate(
            func, *args, engine=engine, engine_kwargs=engine_kwargs, **kwargs
        )

    agg = aggregate

    def nlargest(self, n=5, keep="first"):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(name="nlargest", class_="GroupBy")

    def nsmallest(self, n=5, keep="first"):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(name="nsmallest", class_="GroupBy")

    def unique(self):
        # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        ErrorMessage.method_not_implemented_error(name="unique", class_="GroupBy")

    def apply(self, func, *args, **kwargs):
        # TODO: SNOW-1063349: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
        if not callable(func):
            raise NotImplementedError("No support for non-callable `func`")
        dataframe_result = pd.DataFrame(
            query_compiler=self._query_compiler.groupby_apply(
                self._by,
                agg_func=func,
                axis=self._axis,
                groupby_kwargs=self._kwargs,
                agg_args=args,
                agg_kwargs=kwargs,
                # TODO(https://github.com/modin-project/modin/issues/7096):
                # upstream the series_groupby param to Modin
                series_groupby=True,
            )
        )
        if dataframe_result.columns.equals(pandas.Index([MODIN_UNNAMED_SERIES_LABEL])):
            # rename to the last column of self._df
            # note that upstream modin does not do this yet due to
            # https://github.com/modin-project/modin/issues/7097
            return dataframe_result.squeeze(axis=1).rename(self._df.columns[-1])
        return dataframe_result


def validate_groupby_args(
    by: Any,
    level: Optional[IndexLabel],
    observed: Union[bool, NoDefault],
) -> None:
    """
    Common validation and checks for the groupby arguments that are used by both SeriesGroupBy
    and DataFrameGroupBy.

    Raises:
        TypeError if native pandas series is used as by item, or if both level and by are None
    Warns:
        If observed is True, this parameter is ignored because CategoryDType is not supported with Snowpark pandas API
    """
    # TODO: SNOW-1063350: Modin upgrade - modin.pandas.groupby.SeriesGroupBy functions
    # check if pandas.Series is used as by item, no native pandas series or dataframe
    # object is allowed.
    raise_if_native_pandas_objects(by)
    if not isinstance(by, Series) and is_list_like(by):
        for o in by:
            raise_if_native_pandas_objects(o)

    if level is None and by is None:
        raise TypeError("You have to supply one of 'by' and 'level'")

    if observed is not no_default and observed:
        WarningMessage.ignored_argument(
            operation="groupby",
            argument="observed",
            message="CategoricalDType is not yet supported with Snowpark pandas API, the observed parameter is ignored.",
        )



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import glob
import os
from collections.abc import Hashable
from typing import Any, Callable, Union

import snowflake.snowpark.modin.pandas as pd
from snowflake.snowpark.session import Session

PANDAS_KWARGS = {"names", "index_col", "usecols", "dtype"}


def upload_local_path_to_snowflake_stage(
    session: Session, path: str, sf_stage: str
) -> None:
    """
    Uploads the contents of a local filepath (file or folder) ``path``
    to a staged location ``sf_stage`` in Snowflake.

    Parameters
    ----------
    session : Session
    Session object in Snowpark.

    path : str
    File path to local file or folder.

    sf_stage : str
    Name of Snowflake stage to upload files to.
    """

    # local file that begins with '@' (represents SF stage)
    if path.startswith(r"\@"):
        path = path[1:]

    # Snowflake uses glob patterns by default,
    # so escape any file path for special symbols like *, . in glob patterns
    if os.path.isdir(path):
        files = os.listdir(path)
        files_to_upload = [
            glob.escape(os.path.join(path, file))
            for file in files
            if os.path.isfile(os.path.join(path, file))
        ]
    elif os.path.isfile(path):
        files_to_upload = [glob.escape(path)]
    else:
        raise ValueError("path must be a folder or a file.")

    for file in files_to_upload:
        session.file.put(file, sf_stage, overwrite=True)


def is_local_filepath(filepath: str) -> bool:
    """
    Returns whether a filepath is local.

    Parameters
    ----------
    filepath : str
    File path to file or folder

    Returns
    -------
    bool
    Whether a filepath is local.
    """

    return not filepath.startswith("@") or filepath.startswith(r"\@")


def get_non_pandas_kwargs(kwargs: Any) -> Any:
    """
    Returns a new dict without pandas keyword
    arguments.

    Args:
        kwargs : Dict of keyword arguments to filter.

    Returns:
        dict without pandas kwargs.
    """

    snowpark_reader_kwargs = {
        kwarg_name: kwarg_value
        for kwarg_name, kwarg_value in kwargs.items()
        if kwarg_name not in PANDAS_KWARGS
    }

    return snowpark_reader_kwargs


def get_columns_to_keep_for_usecols(
    usecols: Union[Callable, list[str], list[int]],
    columns: pd.Index,
    maintain_usecols_order: bool = False,
) -> list[Hashable]:
    """
    Returns a subset of `df_columns` to keep, based on `usecols`.

    Parameters
    ----------
    usecols : Callable, list of str, list of int.
        If `usecols` is a Callable, the callable function will be evaluated against the column names,
        returning names where the callable function evaluates to True. If `usecols` is a list, all elements must either
        be positional (i.e. integer indices into the document columns) or strings
        that correspond to column names provided either by the user in `names` or
        inferred from the document header row(s).

    columns : `pd.Index`.
        An index containing all DataFrame column labels

    maintain_usecols_order : bool, default False
        If True, the result's order is based on usecols. Otherwise, the order is based on columns.

    Returns
    -------
    List.
        Subset of columns to keep.

    Raises
    ------
    ValueError
        If column(s) expected in `usecols` are not found in frame's columns `columns`.
    """

    if not isinstance(usecols, list):
        keep = [column for column in columns if usecols(column)]
    else:
        if not usecols:
            keep = []
        else:
            if isinstance(usecols[0], str):
                invalid_columns = [
                    column for column in usecols if column not in columns
                ]
                if invalid_columns:
                    raise ValueError(
                        f"'usecols' do not match columns, columns expected but not found: {invalid_columns}"
                    )
            else:
                invalid_columns = [
                    column for column in usecols if column < 0 or column >= len(columns)  # type: ignore[operator]
                ]
                if invalid_columns:
                    raise ValueError(
                        f"'usecols' do not match columns, columns expected but not found: {invalid_columns}"
                    )

                # Turn index references to pandas labels.
                usecols = [columns[column] for column in usecols]

            l1, l2 = (
                (usecols, columns) if maintain_usecols_order else (columns, usecols)
            )
            keep = [column for column in l1 if column in l2]

    return keep



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import sys
from types import ModuleType
from typing import Any, Callable, Optional

# import the entire context submodule instead of just get_active_session so
# that we can mock get_active_session
import snowflake.snowpark.context
from snowflake.snowpark.exceptions import SnowparkSessionException
from snowflake.snowpark.session import Session, _active_sessions


def _subimport(name: str) -> ModuleType:
    """
    We need this to pickle the session holder class: https://github.com/cloudpipe/cloudpickle/issues/405#issuecomment-756085104
    """
    __import__(name)
    return sys.modules[name]


class SnowpandasSessionHolder(ModuleType):
    """
    This class implements the pattern [1] to make "session" a singleton.

    [1] https://docs.python.org/3.12/reference/datamodel.html#customizing-module-attribute-access
    """

    _session: Optional[Session] = None
    """
    The Snowpark session that Snowpark pandas DataFrame or Series will use.

    It starts as `None`, but if you try to access it when it's `None`:
      - If there is a unique active Snowpark session, snowpark assigns that one to ``session``.
      - If there are no active sessions, or multiple sessions, Snowpark will raise an exception.

    You can assign a value to this session as you would normally assign a
    value to a module property, e.g. `pd.session = session1`.
    """

    def _get_active_session(self) -> Session:
        if self._session is not None and self._session in _active_sessions:
            return self._session

        try:
            session = snowflake.snowpark.context.get_active_session()
            self._session = session
            return session
        except SnowparkSessionException as ex:
            if ex.error_code == "1409":
                raise SnowparkSessionException(
                    "There are multiple active snowpark sessions, but you need to choose one for Snowpark pandas. "
                    + "Please assign one to Snowpark pandas with a statement like `modin.pandas.session = session`."
                ) from ex
            if ex.error_code == "1403":
                raise SnowparkSessionException(
                    "Snowpark pandas requires an active snowpark session, but there is none. Please create one "
                    + "by following the instructions here: https://docs.snowflake.com/en/developer-guide/snowpark/python/creating-session#creating-a-session"
                ) from ex
            raise

    def __setattr__(self, attr: str, value: Any) -> None:
        # If this module is modin.pandas, delegate the attribute to snowflake.snowpark.modin.pandas
        if self.__package__ == "modin.pandas" and attr == "session":
            import snowflake.snowpark.modin.pandas as spd

            setattr(spd, attr, value)
            return
        if attr == "session":
            self._session = value
        else:
            super().__setattr__(attr, value)

    def __getattr__(self, name: str) -> Any:
        # If this module is modin.pandas, delegate the attribute to snowflake.snowpark.modin.pandas
        if self.__package__ == "modin.pandas" and name == "session":
            import snowflake.snowpark.modin.pandas as spd

            return getattr(spd, name)
        return (
            self._get_active_session()
            if name == "session"
            else super().__getattribute__(name)
        )

    def __reduce__(self) -> tuple[Callable[[str], ModuleType], tuple[str]]:
        """
        Implement a custom pickle method so this class is pickleable.

        We need to pickle this class to use the Snowpark pandas module in
        stored procedures.

        Explanation of why we need this to pickle the class: https://github.com/cloudpipe/cloudpickle/issues/405#issuecomment-756085104
        """
        return _subimport, (self.__name__,)
